# encoding: utf-8

# ------------------------------------------------------------------------------
# Copyright (c) 2016 SUSE Linux GmbH, Nuernberg, Germany.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of version 2 of the GNU General Public License as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, contact SUSE Linux GmbH.
#
# ------------------------------------------------------------------------------
#
# Summary: SUSE Connect Program module, main GUI dialog
# Authors: Ilya Manyugin <ilya.manyugin@suse.com>

require 'yast'
require 'suse_connect_program/product'
require 'suse_connect_program/product_controller'

Yast.import 'UI'
Yast.import 'Label'
Yast.import 'Popup'
Yast.import 'Confirm'
Yast.import 'Report'

module SuseConnectProgram
  # Main GUI class
  class MainWindow
    include Yast
    include Yast::UIShortcuts
    include Yast::I18n
    include Yast::Logger

    @search_terms = ''
    @controller = nil

    def initialize(filename, absolute_path)
      textdomain 'suse-connect-program'
      @is_root = Process.uid == 0
      @controller = ProductController.new(filename, absolute_path, @is_root)
    end

    def run
      return unless create_dialog
      begin
        UI.SetFocus(:products_table)
        show_product_details
        return event_loop
      ensure
        close_dialog
      end
    end

    private

    # Create the main dialog window
    def create_dialog
      begin
        @controller.prepare
      rescue ProductControllerGUIError => e
        Popup.Error(e.message)
        return
      end
      UI.OpenDialog(
        Opt(:decorated, :defaultsize),
        dialog_contents
      )
    end

    def close_dialog
      Product.clean_tmp
      UI.CloseDialog
    end

    # Check if running as a priviledged user, show a message if not
    def check_priviledges
      if !@is_root
        Popup.AnyMessage(
          _('Super user priviledges are required.'),
          _("Software management functionality requires super user priviledges.\n"\
          'Please run this module as root before proceeding.'))
        return false
      end
      true
    end

    def event_loop
      loop do
        input = UI.UserInput
        input = input.to_sym if input.is_a?(::String)
        begin
          case input
          when :cancel
            break
          when :products_table
            show_product_details
          when :update_products
            next unless check_priviledges
            begin
              @controller.update_products_list
              refresh_table
            rescue ProductControllerNetworkError => e
              log.error "Could not fetch the latest product list: #{e.message}"
              Popup.Error(_('Could not fetch the latest product list: connection error.'))
            rescue ProductControllerFileNotFound => e
              Popup.Error(_('Could not fetch the latest product list: file doesn\'t'\
                ' exist on the server.'))
              log.error "Could not fetch the latest product list: #{e.message}"
            rescue ProductControllerFileReadError => e
              refresh_table
              Popup.Error(e.message)
            rescue ProductControllerGUIError => e
              refresh_table
              Popup.Warning(e.message)
            else
              Popup.Message(_('Product list was successfully updated.'))
            end
          when :product_install
            install_product if check_priviledges
          when :product_remove
            if check_priviledges
              @controller.remove_resolvable(get_selected_product)
              show_product_details
            end
          when :product_add_repo
            if check_priviledges
              @controller.add_repository(get_selected_product)
              show_product_details
            end
          when :product_rm_repo
            if check_priviledges
              @controller.remove_repository(get_selected_product)
              show_product_details
            end
          when :product_buy
            @controller.open_url(get_selected_product, :buy)
          when :product_doc
            @controller.open_url(get_selected_product, :doc)
          when :product_url
            @controller.open_url(get_selected_product, :product)
          when :product_vendor_url
            @controller.open_url(get_selected_product, :vendor)
          when :product_eula
            @controller.open_url(get_selected_product, :eula)
          when :product_training
            @controller.open_url(get_selected_product, :training)
          when :search_field
            @search_terms = UI.QueryWidget(Id(:search_field), :Value)
            refresh_table
          else
            log.warn "Unexpected input #{input}"
          end
        rescue RuntimeError => e
          Popup.Error(e.message)
        end
      end
    end

    def dialog_contents
      HBox(
        UI.TextMode ? Empty() : HSpacing(5.0),
        VBox(
          Left(Heading(_('SUSE Connect Program'))),
          HBox(
            Left(
              HWeight(1, PushButton(Id(:update_products), _('Re&fresh list')))
            ),
            HStretch(),
            Right(InputField(Id(:search_field), Opt(:hstretch, :notify), _('&Search'), ''))
          ),
          Table(Id(:products_table), Opt(:immediate, :keepSorting),
            Header(_('Product'), _('Vendor'), _('Description')),
            table_entries
          ),
          ReplacePoint(Id(:rp_details), Empty()),
          HBox(
            HStretch(),
            HWeight(1, PushButton(Id(:cancel), Label.QuitButton))
          )
        ),
        UI.TextMode ? Empty() : HSpacing(5.0)
      )
    end

    def install_product
      dependencies_ok, msg = @controller.dependencies_fulfilled?(get_selected_product)
      until dependencies_ok
        ans = Report.AnyQuestion(_('An action is required'), msg, Label.ContinueButton,
          Label.CancelButton, :focus_no)
        return unless ans
        log.error('Opening the addons module')
        @controller.open_addons_manager
        dependencies_ok, msg = @controller.dependencies_fulfilled?(get_selected_product)
      end
      answer = Popup.AnyQuestion(_('Accept License'),
        _('By clicking the Install button you agree with the terms and conditions'\
        ' specified in the End User License Agreement of the product.'),
        _('&Agree'), _('&Decline'), :focus_no)
      return unless answer
      @controller.install_resolvable(get_selected_product)
      show_product_details
    end

    # Show detailed view for the selected product
    def show_product_details
      product = @controller.get_product_details(get_selected_product)
      unless product
        UI.ReplaceWidget(Id(:rp_details), Empty())
        return
      end
      UI.ReplaceWidget(Id(:rp_details), detailed_view(product))
      if product.type == :rpmRepositoryProduct
        if product.repo_added
          UI.ReplaceWidget(Id(:rp_repository), PushButton(Id(:product_rm_repo),
            _('&Remove repository')))
          if product.installed
            UI.ReplaceWidget(Id(:rp_product), PushButton(Id(:product_remove), _('Re&move')))
          else
            UI.ReplaceWidget(Id(:rp_product), PushButton(Id(:product_install), Label.InstallButton))
          end
        else
          UI.ReplaceWidget(Id(:rp_repository), PushButton(Id(:product_add_repo),
            _('&Add repository')))
        end
      end
      UI.ReplaceWidget(Id(:rp_buying), PushButton(Id(:product_buy),
        _('&How to buy'))) if product.how_to_buy
      UI.ReplaceWidget(Id(:rp_documentation), PushButton(Id(:product_doc),
        _('&Documentation'))) if product.doc_url
      UI.ReplaceWidget(Id(:rp_eula), PushButton(Id(:product_eula),
        _('&EULA'))) if product.eula_url
      UI.ReplaceWidget(Id(:rp_training), PushButton(Id(:product_training),
        _('&Trainings'))) if product.training_url
    end

    # Get index of the product selected in the table
    def get_selected_product
      UI.QueryWidget(Id(:products_table), :CurrentItem)
    end

    # Refresh the table
    def refresh_table
      UI.ChangeWidget(Id(:products_table), :Items, table_entries)
      show_product_details
    end

    # Get an array of table entries
    # @return [Array]
    def table_entries
      entries = @controller.filter_product_list(@search_terms)
      entries.map { |entry| Item(Id(entry[0]), *entry[1..-1]) }
    end

    # Show the detailed view for the product
    # @param product [SuseConnectProgram::Product]
    # @return [Yast::UI::Frame]
    def detailed_view(product)
      product_name = if product.installed
                       "#{product.name} (Installed)"
                     else
                       product.name
                     end
      internal = VBox(
        HBox(
          VBox(
            Left(UI.TextMode ? Heading(product_name) : Label(Opt(:boldFont), product_name)),
            Left(Label(product.vendor_name))
          ),
          HStretch(),
          PushButton(Id(:product_url), _('Go to website'))
        ),
        RichText(Opt(:shrinkable), product.product_info),
        Left(
          HBox(
            ReplacePoint(Id(:rp_repository), Empty()),
            ReplacePoint(Id(:rp_product), Empty()),
            ReplacePoint(Id(:rp_buying), Empty()),
            ReplacePoint(Id(:rp_documentation), Empty()),
            ReplacePoint(Id(:rp_training), Empty()),
            ReplacePoint(Id(:rp_eula), Empty())
          )
        )
      )
      UI.TextMode ? Frame('', internal) : internal
    end
  end
end
