# encoding: utf-8

# ------------------------------------------------------------------------------
# Copyright (c) 2016 SUSE Linux GmbH, Nuernberg, Germany.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of version 2 of the GNU General Public License as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, contact SUSE Linux GmbH.
#
# ------------------------------------------------------------------------------
#
# Summary: SUSE Connect Program: products' controller class
# Authors: Ilya Manyugin <ilya.manyugin@suse.com>

require 'fileutils'
require 'net/http'
require 'open-uri'
require 'yast'
require 'suse_connect_program/product'
require 'suse_connect_program/product_manager'

Yast.import 'UI'
Yast.import 'Report'

module SuseConnectProgram
  class ProductController
    
    include Yast::Logger
    include Yast::I18n

    SHARE = '/var/lib/YaST2/suse_connect_program/'.freeze
    REPOSITORY_XML_PATH = 'repodata/repomd.xml'.freeze
    METADATA_PATH = 'metadata.json'.freeze
    
    # Constructor
    # @param filename [String] name of the file
    # @param absolute [Boolean] treat filename as an absolute path
    def initialize(filename, absolute = false, _is_root = true)
      textdomain 'suse-connect-program'
      @products_file = if filename.nil?
                         File.join(SHARE, 'products.json')
                       else
                         absolute ? filename : File.join(SHARE, filename)
                       end
      @product_list = []
    end

    # Obtain the Package Manager lock and read the products file
    def prepare
      @manager = ProductManager.instance
      begin
        @manager.lock
      rescue StandardError => e
        log.error "Error initializing the Package Manager: #{e.message}"
        read_products_file
        raise ProductControllerGUIError, _("Error initializing the Package Manager: #{e.message}")
      end
      read_products_file
    end

    # Update the product_list with the contents of the products file
    def read_products_file
      @product_list = []
      begin
        @product_list = Product.all(@products_file)
      rescue RuntimeError => e
        log.error "Error reading the products file: #{e.message}"
        raise ProductControllerGUIError, e.message
      rescue StandardError => e
        log.error "Error reading the products file: #{e.message}"
        raise ProductControllerGUIError, _('Could not read the products file.')
      end
    end

    # Filter the list of products with respect to the search terms
    # @param search_terms [String]
    def filter_product_list(search_terms)
      products = @product_list.dup
      if search_terms && !search_terms.empty?
        begin
          regex = Regexp.new(search_terms, Regexp::IGNORECASE)
        rescue RegexpError
          return []
        end
        products.select! do |product|
          product.searchable_fields.any? { |f| regex.match(f) }
        end
      end
      products.map(&:table_fields)
    end

    # Retrieve the updated information about the product with the given ID
    # @param product_id [Integer] Product's index in the list
    def get_product_details(product_id)
      begin
        product = get_product_by_id(product_id)
      rescue ProductControllerGUIError
        return nil
      end
      update_product(product) if product.type == :rpmRepositoryProduct
      product
    end

    def update_products_list
      # TODO: so far we only know how to update the rpmRepository products
      not_updated = []
      products = @product_list.select { |p| p.type != :rpmRepositoryProduct }.map(&:json_s)
      url_list = @product_list.select { |p| p.type == :rpmRepositoryProduct }
                              .map(&:repository_url).uniq
      url_list.each do |url|
        md_url = URI.join(url, METADATA_PATH)
        content = uri_get(md_url)
        unless content
          # Copy all the products that have the same repo URL to the list
          affected = @product_list.select { |p| p.repository_url == url }
          products.concat(affected.map(&:json_s))
          not_updated.concat(affected.map(&:name))
          next
        end
        begin
          j = JSON.parse(content.join, symbolize_names: true)
          j.each do |product|
            if product[:product_logo].nil?
              product[:product_logo] = ''
              next
            end
            next unless product[:product_logo] =~ URI.regexp
            logo = uri_get(URI(product[:product_logo]))
            product[:product_logo] = if logo.nil?
                                       ''
                                     else
                                       logo = logo.each(&:strip).join
                                       Base64.encode64(logo).strip
                                     end
          end
          products.concat(j)
        rescue JSON::ParserError => e
          log.error "Error parsing metadata at URL '#{md_url}': #{e.message}"
          affected = @product_list.select { |p| p.repository_url == url }
          products.concat(affected.map(&:json_s))
          not_updated.concat(affected.map(&:name))
          next
        end
      end
      if write_products_to_file(products) && !not_updated.empty?
        not_updated_s = not_updated.join(', ')
        raise ProductControllerGUIError,
          _("The following products could not be updated: #{not_updated_s}.")
      end
    end

    def write_products_to_file(products)
      new_file_name = @products_file + '.new'
      backup_file_name = @products_file + '.bak'
      begin
        File.open(new_file_name, 'w') do |f|
          f.write(JSON.pretty_generate(products))
        end
      rescue StandardError => e
        log.error "Could not write the product list to a file: #{e.message}"
        raise ProductControllerGUIError, _('Could not write the new products file to disk.')
      end
      FileUtils.mv(@products_file, backup_file_name)
      FileUtils.mv(new_file_name, @products_file)
      begin
        read_products_file
      rescue ProductControllerGUIError
        log.error 'Fetch latest product list: The new file could not be read. Switching to backup.'
        FileUtils.mv(backup_file_name, @products_file)
        read_products_file
        raise ProductControllerFileReadError,
          _('Reading the new products file failed. Switching back to the old one.')
      end
    end

    # Add the repository for the product with given ID
    # @param product_id [Integer] Product's index in the list
    def add_repository(product_id)
      product = get_product_by_id(product_id)
      check_repo_accessible(product.repository_url)
      begin
        result = @manager.add_repository(product.repository_url,
          product.repo_name, product.vendor_name)
      rescue SuseConnectProgram::ProductManagerException => e
        raise ProductControllerGUIError, e.message
      end
      update_product(product)
      result
    end

    # Install the resolvable for the product with the given ID
    # @param product_id [Integer] Product's index in the list
    def install_resolvable(product_id)
      product = get_product_by_id(product_id)
      result = @manager.install_resolvable(product.resolvable, product.repository_url)
      update_product(product)
      result
    end

    # Remove the repository of the product with the given ID
    # @param product_id [Integer] Product's index in the list
    def remove_repository(product_id)
      product = get_product_by_id(product_id)
      result = @manager.remove_repository(product.repository_url)
      update_product(product)
      result
    end

    # Remove the resolvable of the product with the given ID
    # @param product_id [Integer] Product's index in the list
    def remove_resolvable(product_id)
      product = get_product_by_id(product_id)
      result = @manager.remove_resolvable(product.resolvable, product.repository_url)
      update_product(product)
      result
    end

    # Open the URL of the specific kind for the product with the given ID
    # @param product_id [Integer] Product's index in the list
    # @param url_kind [Symbol] Kind of the URL
    def open_url(product_id, url_kind)
      product = get_product_by_id(product_id)
      url = case url_kind
            when :buy
              product.how_to_buy
            when :doc
              product.doc_url
            when :product
              product.product_url
            when :vendor
              product.vendor_url
            when :eula
              product.eula_url
            when :training
              product.training_url
            else
              ''
            end
      if url.empty? || url.nil?
        raise _("There is no #{url_kind} URL specified for product #{product_id}.")
      end
      if Yast::UI.TextMode
        `which w3m`
        if $?.success?
          Yast::UI.RunInTerminal("w3m #{url}")
        else
          raise _('The console-based web browser w3m is not installed. Cannot open the webpage.')
        end
      else
        `which xdg-open`
        if $?.success?
          system("xdg-open #{url}")
        else
          raise _('The xdg-utils package is not installed. Cannot open the webpage.')
        end
      end
    end

    def dependencies_fulfilled?(product_id)
      product = get_product_by_id(product_id)
      return [true, ''] if product.dependencies.nil? || product.dependencies.empty?
      unmet = []
      str = ''
      product.dependencies.each do |dep|
        if !@manager.dependency_met?(dep[:resolvable_type], dep[:resolvable_name])
          unmet << dep[:display_name]
        end
      end
      if !unmet.empty?
        str << _("The following module dependencies are not installed in the system:\n")
        str << unmet.map { |e| "- #{e}\n" }.join
        str << _('The Add-Ons Manager will open now. Would you like to continue?')
      end
      [unmet.empty?, str]
    end

    def open_addons_manager
      if !Yast::WFM.ClientExists('inst_scc')
        raise ProductControllerGUIError,
          _('Could not open the YaST Add-Ons Manager: the module is not installed.')
      end
      Yast.import 'Wizard'
      Yast::Wizard.CreateDialog
      Yast::WFM.CallFunction('inst_scc', ['select_extensions'])
      Yast::Wizard.CloseDialog
    end

    private

    # Get the Product instance by its ID
    # @param product_id [Integer]
    def get_product_by_id(product_id)
      product = @product_list.find { |prod| prod.id == product_id }
      raise ProductControllerGUIError, _('Incorrect selection.') unless product
      product
    end

    # Update product fields
    # @param product[SuseConnectProgram::Product]
    def update_product(product)
      product.update(@manager.repository_added?(product.repository_url),
        @manager.resolvable_installed?(product.resolvable))
    end

    # Fetch the file contents through a URI
    # @param uri [URI] URI
    def uri_get(uri)
      retries = 3
      begin
        uri.open("User-Agent" => "SUSE Connect Program Client", redirect: false) { |f| f.readlines }
      rescue OpenURI::HTTPRedirect => e
        # handle redirect manually
        retries -= 1
        uri = e.uri
        retry unless retries < 0
        nil
      rescue OpenURI::HTTPError => e
        log.error "Error fetching URL '#{uri}': #{e.message}"
        nil
      rescue NoMethodError => e # URI is invalid
        log.error "Error fetching URL '#{uri}': URI is invalid"
        nil
      end
    end

    # Check if the URL is accessible
    # @param url [String]
    def check_repo_accessible(url)
      log.info "--- #{self.class}.#{__callee__}(url) ---"
      local_url = url.dup
      if !local_url.end_with? REPOSITORY_XML_PATH
        local_url << '/' unless local_url.end_with? '/'
        local_url << REPOSITORY_XML_PATH
      end
      uri = URI(local_url)
      begin
        response = Net::HTTP.get_response(uri)
      rescue SocketError, Errno::ENETUNREACH, Exception => e
        log.error("Could not connect to #{uri.host}. Error: #{e.inspect}")
        raise ProductControllerGUIError,
          _("Could not connect to #{uri.host}. Check your network connection.")
      end
      # Handle the redirect code 302 -- "found"
      unless response.code == '200' || response.code == '302'
        log.error "Could not find the repository on the remote server:"\
          " #{response.uri}, code #{response.code}."
        raise ProductControllerGUIError, _('Could not find the repository on the remote server.')
      end
      true
    end
  end

  # Raised if the URL could not be reached
  class ProductControllerNetworkError < RuntimeError
  end

  # Raised if the products file is not found
  class ProductControllerFileNotFound < RuntimeError
  end

  # Raised if the newly fetched products file could not be read and parsed
  class ProductControllerFileReadError < RuntimeError
  end

  # Raised in order to show a GUI message
  class ProductControllerGUIError < RuntimeError
  end

end
