# encoding: utf-8

# ------------------------------------------------------------------------------
# Copyright (c) 2016 SUSE Linux GmbH, Nuernberg, Germany.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of version 2 of the GNU General Public License as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, contact SUSE Linux GmbH.
#
# ------------------------------------------------------------------------------
#
# Summary: SUSE Connect Program module, product manager
#          The product manager adds the necessary repositories and installs resolvables.
# Authors: Ilya Manyugin <ilya.manyugin@suse.com>

require 'yast'
require 'net/http'
require 'singleton'

module SuseConnectProgram
  Yast.import 'Pkg'
  Yast.import 'Package'
  Yast.import 'PackageLock'
  Yast.import 'KeyManager'
  Yast.import 'PackageCallbacks'
  Yast.import 'SignatureCheckCallbacks'

  class ProductManagerException < Exception
  end

  # ProductManager: class responsible for managing repositories and software packages
  class ProductManager
    include Singleton

    include Yast
    include Yast::Logger
    include Yast::I18n

    def initialize
      textdomain 'suse-connect-program'
      @sources = []
    end

    def lock
      lock = PackageLock.Connect(false)
      raise_pkg_exception unless lock['connected']
      raise_pkg_exception unless Pkg.TargetInitialize('/')
      raise_pkg_exception unless Pkg.TargetLoad
      Pkg.SourceStartManager(true)
      PackageCallbacks.InitPackageCallbacks
      Pkg.CallbackImportGpgKey(FunRef.new(SignatureCheckCallbacks
        .method(:import_gpg_key_or_disable),
        'boolean (map <string, any>, integer)'))
      KeyManager.Read
      refresh_sources
    end

    # Add repository by its URL to the system
    # @param url [String] repository url
    # @param repo_name [String] vendor name with non-letters substituted by underscores
    # @param vendor_name [String]
    # @return [Boolean] whether or not the addition was successful
    def add_repository(url, repo_name, vendor_name)
      return true if repository_added?(url)
      log.info("Adding repository '#{url}' for vendor '#{vendor_name}'.")
      repo_type = Pkg.RepositoryProbe(url, '')
      log.info("Probed repository: repository type is '#{repo_type}'.")
      if repo_type != 'YUM'
        log.error("Incorrect repository type for vendor '#{vendor_name}': "\
          "type is '#{repo_type}', expected 'YUM'.")
        raise ProductManagerException, _("Incorrect repository type: #{repo_type}")
      end
      repo_properties = {
        'enabled' => true,
        'autorefresh' => true,
        'name' => vendor_name,
        'prod_dir' => '/',
        'alias' => repo_name,
        'base_urls' => [url],
        'type' => repo_type
      }
      source_id = Pkg.RepositoryAdd(repo_properties)
      raise ProductManagerException, _('Could not add the repository') unless source_id
      log.debug("Repository added. Source ID is #{source_id}.")
      general_data = Pkg.SourceGeneralData(source_id)
      log.debug("SourceGeneralData: #{general_data}")
      success = Pkg.SourceRefreshNow(source_id)
      unless success
        log.error("SourceRefreshNow returned #{success} on source ID #{source_id}")
      end
      KeyManager.Write
      Pkg.SourceSaveAll
      refresh_sources
      src = @sources.find { |s| s['SrcId'] == source_id }
      if src.nil? or !src['enabled']
        # repository was disabled by the callback
        remove_repository(url)
        raise ProductManagerException,
          _('The repository key was not accepted. The repository was not added.')
      end
      success
    end

    # Remove repository from the system by its URL
    # @param url [String]
    def remove_repository(url)
      source_id = source_for_url(url)
      unless source_id
        log.warn("Repository was not added for URL '#{url}', yet remove_repository was called")
        return true
      end
      result = Pkg.SourceDelete(source_id)
      Pkg.SourceSaveAll
      refresh_sources
      result
    end

    # Check if the repository for the product is added to the system
    # @param repository_url [String]
    # @return [Boolean]
    def repository_added?(repository_url)
      src_id = source_for_url(repository_url)
      src_id ? true : false
    end

    # Install the resolvable from the product's repository
    # @param resolvable_name [String]
    # @param repository_url [String]
    def install_resolvable(resolvable_name, repository_url)
      resolvable_kind, resolvable_props = find_resolvables(resolvable_name)
      result = true
      if resolvable_props.empty?
        log.error("Could not find an appropriate resolvable '#{resolvable_name}'")
        raise _("Could not find an appropriate resolvable '#{resolvable_name}'")
      end
      target_source = source_for_url(repository_url)
      res_prop = resolvable_props.find { |prop| prop['source'] == target_source }
      unless res_prop
        log.error("Resolvable #{resolvable_name} found in the repositories, "\
          "but not in the product's repository.")
        raise _("Could not find '#{resolvable_name}' in the product's repository "\
          "(#{product_name}, #{target_source})")
      end
      if res_prop['status'] == :installed
        log.warn("Resolvable #{resolvable_name} is already installed from the source #{source_id}.")
        return true
      elsif res_prop['status'] == :available
        result = Pkg.ResolvableInstallRepo(resolvable_name, resolvable_kind, target_source)
        result &&= Pkg.PkgSolve(false)
        WFM.CallFunction('inst_rpmcopy')
      else
        raise _("Resolvable #{resolvable_name} is in status #{res_prop["status"]}")
      end
      result
    end

    # Check if the product's resolvable is installed in the system
    # @param resolvable_name [String]
    # @return [Boolean]
    def resolvable_installed?(resolvable_name)
      # The installed resolvable will always have "source" => -1 in them, so there is no point
      # in finding the resolvable with "source" == product's source
      _, resolvables = find_resolvables(resolvable_name)
      if resolvables.empty?
        log.error("Resolvable '#{resolvable_name}' could not be found in the repositories.")
        return false
      end
      resolvable = resolvables.find { |resolv| resolv['source'] == -1 }
      !resolvable.nil? && resolvable['status'] == :installed
    end

    # Remove resolvable from the system
    # @param resolvable_name [String]
    # @param repository_url [String]
    # @return [Boolean]
    def remove_resolvable(resolvable_name, repository_url)
      resolvable_kind, resolvable = find_resolvable_by_url(resolvable_name, repository_url)
      unless resolvable
        log.error("Resolvable '#{resolvable_name}' could not be found in the repository"\
        " for product '#{product_name}'.")
        return true
      end
      result = Pkg.ResolvableRemove(resolvable_name, resolvable_kind)
      result &&= Pkg.PkgSolve
      WFM.CallFunction('inst_rpmcopy')
      result
    end

    # List all added repositories
    def list_repositories
      @sources.each { |source| puts "#{source['SrcId']}: #{source['name']}" }
    end

    # Check if a dependency is met
    def dependency_met?(resolvable_type, resolvable_name)
      properties = Pkg.ResolvableProperties(resolvable_name, resolvable_type.to_sym, '')
      !properties.empty?
    end

    protected

    def raise_pkg_exception(message = nil)
      if message
        log.error(message)
      else
        message = Pkg.LastError
      end
      raise message
    end

    # Find resolvable (pattern or package) in the repository belonging to the product with
    # the specified name
    # @param resolvable_name [String]
    # @param product_name [String]
    def find_resolvable_in_product(resolvable_name, product_name)
      source_id = source_for_product(product_name)
      resolvable_kind, resolvable_props = find_resolvables(resolvable_name)
      resolvable = resolvable_props.find  { |resolv| resolv['source'] == source_id }
      [resolvable_kind, resolvable]
    end

    # Find resolvable in the repository specified by its URL
    # @param resolvable_name [String]
    # @param repository_url [String]
    def find_resolvable_by_url(resolvable_name, repository_url)
      source_id = source_for_url(repository_url)
      resolvable_kind, resolvable_props = find_resolvables(resolvable_name)
      resolvable = resolvable_props.find  { |resolv| resolv['source'] == source_id }
      [resolvable_kind, resolvable]
    end

    # Find all resolvables of types 'pattern' and 'package' in all the repositories registered
    # in the system
    # @param resolvable_name [String]
    # @return [resolvable_kind, resolvable_props] [[Symbol, nil], [Array]]
    def find_resolvables(resolvable_name)
      resolvable_props = nil
      resolvable_kind = nil
      [:pattern, :package].each do |type|
        resolvable_props = Pkg.ResolvableProperties(resolvable_name, type, '')
        unless resolvable_props.nil? || resolvable_props.empty?
          resolvable_kind = type
          break
        end
      end
      [resolvable_kind, resolvable_props]
    end

    # Update the internal source list
    def refresh_sources
      Pkg.SourceLoad
      @sources = Pkg.SourceEditGet.each do |source|
        source[:data] = Pkg.SourceGeneralData(source.fetch('SrcId', -1))
      end
    end

    # Find the source having the specified base URL
    # @param url [String]
    # @return [Integer, nil] source ID or nil if not added
    def source_for_url(url)
      begin
        src = @sources.find do |source|
          source[:data]['base_urls'].any? { |burl| burl == url || burl == url + '/' }
        end
        src ? src['SrcId'] : nil
      rescue Exception
        nil
      end
    end
  end
end
